﻿using UnityEngine;
using System.Collections;

public class PlayerController : MonoBehaviour 
{
    [SerializeField] Animator playerAnimator = null;

    [SerializeField] float gravity = 9.81f;
    [SerializeField] float maxRunSpeed = 5.0f;
    [SerializeField] float runAcceleration = 10.0f;
    [SerializeField] float friction = 100.0f;
    [SerializeField] float jumpHeight = 4.0f;

    [SerializeField] float dashSpeed = 20.0f;
    [SerializeField] float dashDuration = 0.5f;
    [SerializeField] float dashInterruptableAfter = 0.4f;
    [SerializeField] AnimationCurve dashCurve = AnimationCurve.Linear(0,1,1,0);
    [SerializeField] int dashCount = 2;
    [SerializeField] GameObject dashEffect = null;
    [SerializeField] Transform dashAttachPoint = null;

    [SerializeField] Transform impactOrigin = null;
    [SerializeField] Transform impactPoint = null;
    [SerializeField] float impactDistance = 1.5f;
    
    [SerializeField] float groundedCastSize = 1.0f;
    [SerializeField] float groundedCastDist = 0.1f;
    [SerializeField] LayerMask groundedCastLayers;

    [SerializeField] float sideCastHeight = 1.0f;

    [SerializeField] AudioClip jumpClip;
    [SerializeField] AudioClip dashClip;
    [SerializeField] AudioClip attackWindupClip;

    // Precalculated
    private float jumpVelocity;

    // State
    private bool grounded = false;
    private bool leftClear = false;
    private bool rightClear = false;
    private int dashesAvailable = 0;
    private bool dashing = false;
    private float dashStartTime = 0.0f;
    private float attackAvailable = 0.0f;
    private Vector3 dashDirection = Vector3.right;
    private Vector3 attackDirection = Vector3.right;

    // Input state
    private bool dashDown = false;
    private bool dashQueued = false;
    private bool attackQueued = false;


    private bool AttackInProgress
    {
        get 
        {
            AnimatorStateInfo infoA = playerAnimator.GetCurrentAnimatorStateInfo(0);
            AnimatorStateInfo infoB = playerAnimator.GetNextAnimatorStateInfo(0);
            return infoA.IsTag("Attack") || infoB.IsTag("Attack");
        }
    }
    private bool DashInProgress
    {
        get { return dashing; }
    }
    private bool AttackBlocked
    {
        get
        {
            if(dashing && (Time.time - dashStartTime) < dashInterruptableAfter)
                return true;
            else
                return false;
        }
    }


    void Awake()
    {
        TrackPlayer.target = impactOrigin;
        GameManager.Instance.InitializeRound(); // Ensure it's built

        jumpVelocity = Mathf.Sqrt(2 * gravity * jumpHeight);
    }

    void FixedUpdate()
    {
        CheckGrounding();
        ApplyGravity();

        HandleInput();

        UpdateAnimator();
    }

    private void CheckGrounding()
    {
        if (dashing)
        {
            grounded = false;
            return;
        }

        LayerMask castLayers = groundedCastLayers;
        float raycastFudge = 0.1f; // To prevent penetration at the start from causing grounded tests to fail

        // Raycast for grounding
        RaycastHit hit;
        grounded = Physics.SphereCast(transform.position + Vector3.up * (groundedCastSize + raycastFudge), groundedCastSize, -Vector3.up, out hit, raycastFudge + groundedCastDist, castLayers);

        leftClear = !Physics.SphereCast(transform.position + Vector3.up * sideCastHeight + Vector3.right * raycastFudge, groundedCastSize, -Vector3.right, out hit, raycastFudge + groundedCastDist, castLayers);
        rightClear = !Physics.SphereCast(transform.position + Vector3.up * sideCastHeight - Vector3.right * raycastFudge, groundedCastSize, Vector3.right, out hit, raycastFudge + groundedCastDist, castLayers);

        if (grounded)
            dashesAvailable = dashCount;
    }

    private void ApplyGravity()
    {
        if(!dashing)
            rigidbody.AddForce(-Vector3.up * gravity, ForceMode.Acceleration);
    }

    private float lookAtDampTime = 0.05f;
    private Vector3 lookAtDampVelocity = Vector3.zero;

    private void HandleInput()
    {
        float horizontal = Input.GetAxisRaw("Horizontal");
        float vertical = Input.GetAxisRaw("Vertical");

        if (Mathf.Abs(horizontal) > .1f || Mathf.Abs(vertical) > .1f) // If no controls, use last given direction
        {
            if (!DashInProgress)
            {
                dashDirection = new Vector3(horizontal, vertical).normalized;
            }
            if (!AttackInProgress)
            {
                attackDirection = new Vector3(horizontal, vertical).normalized;
            }
        }
        else
        {
            if (!AttackInProgress)
            {
                attackDirection = Mathf.Sign(transform.forward.x) * Vector3.right;
            }
        }

        //if (!jumpDown && vertical > 0.5f)
        //{
        //    OnJumpButtonDown();
        //}
        //jumpDown = vertical > 0.5f;

        float dash = Input.GetAxisRaw("Dash");
        if (!dashDown && dash > 0.0f)
        {
            if (DashInProgress)
            {
                dashQueued = true;
            }
            if(!DashInProgress)
            {
                if (grounded)
                    OnJumpButtonDown();
                else
                    OnDashButtonDown();
            }
        }
        dashDown = dash > 0.0f;

        if (Input.GetAxisRaw("Fire1") > 0.0f)
        {
            if (!AttackInProgress && !AttackBlocked)
            {
                OnAttackButtonDown();
            }
            else if (AttackBlocked)
            {
                attackQueued = true;
            }            
        }

        if (!DashInProgress)
        {
            if (dashQueued)
            {
                OnDashButtonDown();
                dashQueued = false;
            }
        }
        if(!AttackInProgress && !AttackBlocked && attackQueued)
        {
            OnAttackButtonDown();
            attackQueued = false;
        }
        if (dashing)
        {
            HandleDash();
        }
        if (AttackInProgress)
        {
            HandleAttack();
        }        
        if(!DashInProgress)
        {
            HandleHorizontalMovement();
        }
    }

    private void HandleHorizontalMovement()
    {
        float horizontal = Input.GetAxisRaw("Horizontal");

        Vector3 horizontalAcceleration = horizontal * Vector3.right * runAcceleration;
        if (Mathf.Abs(rigidbody.velocity.x) < maxRunSpeed || Mathf.Sign(rigidbody.velocity.x) != Mathf.Sign(horizontalAcceleration.x))
        {
            if ((horizontalAcceleration.x < 0 && leftClear) || (horizontalAcceleration.x > 0 && rightClear))
            {
                rigidbody.AddForce(horizontalAcceleration);
            }
        }

        // Handle orientation
        Vector3 targetLookAt;
        if (!AttackInProgress && grounded && Mathf.Abs(horizontal) > .5f)
        {
            targetLookAt = new Vector3(Mathf.Sign(horizontal), 0.0f, -.01f);
        }
        else
        {
            targetLookAt = Mathf.Sign(transform.forward.x) * Vector3.right + new Vector3(0, 0, -.01f);
        }

        Vector3 dampForward = Vector3.SmoothDamp(transform.forward, targetLookAt, ref lookAtDampVelocity, lookAtDampTime);
        transform.LookAt(transform.position + dampForward);

        if (Mathf.Abs(horizontal) < .1f && grounded && Mathf.Abs(rigidbody.velocity.x) > 0.5f)
        {
            float frictionForce = -Mathf.Sign(rigidbody.velocity.x) * friction;

            rigidbody.AddForce(new Vector3(frictionForce, 0));
        }
    }

    private void OnJumpButtonDown()
    {
        if (grounded && !dashing)
        {
            Vector3 vel = rigidbody.velocity;
            vel.y = jumpVelocity;
            rigidbody.velocity = vel;

            audio.PlayOneShot(jumpClip);
        }
    }

    private void OnDashButtonDown()
    {
        if (dashesAvailable > 0 && !dashing)
        {
            dashing = true;
            dashStartTime = Time.time;

            if (dashEffect)
            {
                GameObject effect = Instantiate(dashEffect) as GameObject;

                effect.transform.parent = dashAttachPoint;
                effect.transform.localPosition = Vector3.zero;
            }

            audio.PlayOneShot(dashClip);

            dashesAvailable--;
        }
    }

    private void HandleDash()
    {
        // During a dash we want to manually set velocity the entire time!
        float timeAlong = (Time.time - dashStartTime) / dashDuration;

        if(timeAlong >= 1.0f)
        {
            dashing = false;
            return;
        }

        float speed = dashCurve.Evaluate(timeAlong) * dashSpeed;
        rigidbody.velocity = dashDirection * speed;
    }

    private void OnAttackButtonDown()
    {
        StartCoroutine(SendAnimatorEvent("AttackEvent"));

        if (Time.time > attackAvailable)
        {
            audio.PlayOneShot(attackWindupClip, 0.5f);
            attackAvailable = Time.time + 0.5f;
        }
    }

    private void HandleAttack()
    {
        impactPoint.position = impactOrigin.position + attackDirection * impactDistance;
        impactPoint.LookAt(impactPoint.position + attackDirection);
    }

    private void UpdateAnimator()
    {
        //STATES:
        //HorizontalDirection
        //VerticalDirection
        //Grounded
        //Dashing
        //Attacking

        playerAnimator.SetBool("Grounded", grounded);
        playerAnimator.SetBool("Dashing", dashing);

        playerAnimator.SetFloat("HorizontalDirection", rigidbody.velocity.x / maxRunSpeed * Mathf.Sign(transform.forward.x));
        playerAnimator.SetFloat("VerticalDirection", rigidbody.velocity.y / jumpVelocity);
        if(DashInProgress)
        {
            playerAnimator.SetFloat("ActionX", dashDirection.x * Mathf.Sign(transform.forward.x));
            playerAnimator.SetFloat("ActionY", dashDirection.y);
        }
        if(AttackInProgress)
        {
            playerAnimator.SetFloat("ActionX", attackDirection.x * Mathf.Sign(transform.forward.x));
            playerAnimator.SetFloat("ActionY", attackDirection.y);
        }
    }

    IEnumerator SendAnimatorEvent(string property)
    {
        playerAnimator.SetBool(property, true);
        yield return null;
        playerAnimator.SetBool(property, false);
    }
}
